@@ -1,6 +1,6 @@
 /* cbmplugs -- Plugins to enable handling of Commodore 64 images in The GIMP
  *
- * Copyright (C) 2002, 2003, 2004 David Weinehall <tao@acc.umu.se>
+ * Copyright (C) 2002-2006 David Weinehall <tao@acc.umu.se>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
  */
 
 #include <stdio.h>
@@ -236,6 +236,12 @@ static guchar vicpalette[] = {
 #define SHFXL_BITMAP	(0x6000 - SHFXL_BASE + 2)
 #define SHFXL_SIZE	(0x7c00 - SHFXL_BASE + 2)
 
+#define HWS_BASE	0x4000
+#define HWS_BITMAP	(0x4000 - HWS_BASE + 2)
+#define HWS_SPRMAP1	(0x5f40 - HWS_BASE + 2)
+#define HWS_SPRMAP2	(0x6840 - HWS_BASE + 2)
+#define HWS_SIZE	(0x7140 - HWS_BASE + 2)
+
 #define swap(_x, _y)	\
 	do { gint _t; _t = _x; _x = _y; _y = _t; } while (0)
 
@@ -358,7 +364,7 @@ static guint8 data[65536];	/* Since the Commodore 64 only has 64kB
 				 * of memory, it is reasonable to assume
 				 * that no image is larger than this...
 				 */
-static gchar *buf;
+static guchar *buf;
 
 static gint32 rain_to_gimp(void)
 {
@@ -375,7 +381,7 @@ static gint32 rain_to_gimp(void)
 	guint8 colour1;
 	guint8 colour2;
 	guint8 x, y;
-	gchar *bufp;
+	guchar *bufp;
 	guint width = 128, height = 128;
 
 	if (baseaddr != RAIN_BASE)
@@ -389,8 +395,8 @@ static gint32 rain_to_gimp(void)
 	 * does not contain any colour information; hence
 	 * we default to the colours used in Rainmaker
 	 */
-	if (data[RAIN_MAGIC] != 0x52 || data[RAIN_MAGIC + 1] != 0x41 ||
-	    data[RAIN_MAGIC + 2] != 0x49 || data[RAIN_MAGIC + 3] != 0x4e) {
+	if (data[RAIN_MAGIC] != 0x52 || data[RAIN_MAGIC + 1] != 0x1 ||
+	    data[RAIN_MAGIC + 2] != 0x9 || data[RAIN_MAGIC + 3] != 0x4e) {
 		colour0 = 0xb;
 		colour1 = 0xc;
 		colour2 = 0xf;
@@ -414,7 +420,7 @@ static gint32 rain_to_gimp(void)
 	gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, width, height,
 			    TRUE, FALSE);
 
-	buf = g_new0(guint8, width * height);
+	buf = g_new0(guchar, width * height);
 
 	/* layer 0: sprmap2 */
 	bufp = buf + 128; /* We begin at (0, 1) rather than (0, 0)
@@ -425,7 +431,7 @@ static gint32 rain_to_gimp(void)
 		for (x = 0; x < 8; x++) {
 			guint i, tmp;
 
-			tmp = sprmap2[x % 3 + (x / 3) * 64 + rainoff2[y]];
+			tmp = sprmap2[x % 3 + (x / 3) * 6 + rainoff2[y]];
 
 			for (i = 0; i < 8; i++) {
 				*bufp++ = (tmp & 1 << (7 - i)) ? colour2 : background;
@@ -435,7 +441,7 @@ static gint32 rain_to_gimp(void)
 	}
 
 	/* layer 1: sprmap1 */
-	bufp = buf + 132; /* We begin at (4, 1) rather than (0, 0)
+	bufp = buf + 132; /* We begin at (, 1) rather than (0, 0)
 			   * for this layer
 			   */
 
@@ -443,7 +449,7 @@ static gint32 rain_to_gimp(void)
 		for (x = 0; x < 15; x++) {
 			guint i, tmp;
 
-			tmp = sprmap1[x % 3 + (x / 3) * 64 + rainoff1[y]];
+			tmp = sprmap1[x % 3 + (x / 3) * 6 + rainoff1[y]];
 
 			for (i = 0; i < 8; i++, bufp++)
 				if (tmp & 1 << (7 - i))
@@ -530,6 +536,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + ADVOCP_COLOUR;
 		background = data + ADVOCP_BG;
 		break;
+
 	case AFLI:
 		if (baseaddr != AFLI_BASE)
 			return -1;
@@ -546,6 +553,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colmode = FLI8;
 		mode = HIRES;
 		break;
+
 	case ARTIST:
 		if (baseaddr != ARTIST_BASE)
 			return -1;
@@ -555,6 +563,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + ARTIST_COLOUR;
 		background = data + ARTIST_BG;
 		break;
+
 	case BLAZE:
 		if (baseaddr != BLAZE_BASE)
 			return -1;
@@ -564,6 +573,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + BLAZE_COLOUR;
 		background = NULL;
 		break;
+
 	case DRAZ:
 		if (baseaddr != DRAZ_BASE)
 			return -1;
@@ -573,6 +583,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + DRAZ_COLOUR;
 		background = data + DRAZ_BG;
 		break;
+
 	case EAFLI:
 		if (baseaddr != EAFLI_BASE)
 			return -1;
@@ -599,6 +610,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		mode = HIRES;
 		height = 84;
 		break;
+
 	case FLI:
 		if (baseaddr != FLI_BASE &&
 		    baseaddr != FLIV2_BASE)
@@ -631,6 +643,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		}
 		colmode = FLI8;
 		break;
+
 	case IMSYS:
 		if (baseaddr != IMSYS_BASE)
 			return -1;
@@ -640,6 +653,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + IMSYS_COLOUR;
 		background = NULL;
 		break;
+
 	case IP64H:
 		if (baseaddr != IP64H_BASE)
 			return -1;
@@ -648,6 +662,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		mcolour0 = data + IP64H_MCOLOUR;
 		mode = HIRES;
 		break;
+
 	case IP64L:
 		if (baseaddr != IP64L_BASE)
 			return -1;
@@ -657,6 +672,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + IP64L_COLOUR;
 		background = data + IP64L_BG;
 		break;
+
 	case KOALA:
 		if (baseaddr != KOALA_BASE)
 			return -1;
@@ -666,6 +682,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + KOALA_COLOUR;
 		background = data + KOALA_BG;
 		break;
+
 	case OCP:
 		if (baseaddr != OCP_BASE)
 			return -1;
@@ -674,6 +691,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		mcolour0 = data + OCP_MCOLOUR;
 		mode = HIRES;
 		break;
+
 	case SHFLI:
 		if (baseaddr != SHFLI_BASE)
 			return -1;
@@ -696,6 +714,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		width = 96;
 		height = 168;
 		break;
+
 	case SHFXL:
 		if (baseaddr != SHFXL_BASE)
 			return -1;
@@ -717,18 +736,21 @@ static gint32 bitmap_to_gimp(guint8 type)
 		width = 144;
 		height = 168;
 		break;
+
 	case SUXXV2:
 		bitmap = data + SUXXV2_BITMAP;
 		mcolour0 = data + SUXXV2_MCOLOUR;
 		colour = data + SUXXV2_COLOUR;
 		background = NULL;
 		break;
+
 	case SUXXV3:
 		bitmap = data + SUXXV3_BITMAP;
 		mcolour0 = data + SUXXV3_MCOLOUR;
 		colour = data + SUXXV3_COLOUR;
 		background = data + SUXXV3_BG;
 		break;
+
 	case TLIGHT:
 		if (baseaddr != TLIGHT_BASE)
 			return -1;
@@ -738,6 +760,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + TLIGHT_COLOUR;
 		background = data + TLIGHT_BG;
 		break;
+
 	case VIDCOM:
 		if (baseaddr != VIDCOM_BASE)
 			return -1;
@@ -747,6 +770,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 		colour = data + VIDCOM_COLOUR;
 		background = NULL;
 		break;
+
 	default:
 		return -1;
 	}
@@ -756,7 +780,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 
 		for (i = 0; i < 500; i++) {
 			tmpcol[i] = colour[i] & 0xf;
-			tmpcol[i + 500] = colour[i] >> 4& 0xf;
+			tmpcol[i + 500] = colour[i] >> 4 & 0xf;
 		}
 		colour = tmpcol;
 	}
@@ -773,7 +797,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 	gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, width, height,
 			    TRUE, FALSE);
 
-	buf = g_new0(guint8, width * height);
+	buf = g_new0(guchar, width * height);
 
 	bufp = buf + 8;
 
@@ -792,24 +816,31 @@ static gint32 bitmap_to_gimp(guint8 type)
 			case 0:
 				cp = mcolour0;
 				break;
+
 			case 1:
 				cp = mcolour1;
 				break;
+
 			case 2:
 				cp = mcolour2;
 				break;
+
 			case 3:
 				cp = mcolour3;
 				break;
+
 			case 4:
 				cp = mcolour4;
 				break;
+
 			case 5:
 				cp = mcolour5;
 				break;
+
 			case 6:
 				cp = mcolour6;
 				break;
+
 			case 7:
 				cp = mcolour7;
 				break;
@@ -843,9 +874,11 @@ static gint32 bitmap_to_gimp(guint8 type)
 					case 0:
 						tmp = col2;
 						break;
+
 					case 1:
 						tmp = col1;
 						break;
+
 					default:
 						break;
 					}
@@ -858,15 +891,19 @@ static gint32 bitmap_to_gimp(guint8 type)
 							*(background + off) &
 							0xf;
 						break;
+
 					case 1:
 						tmp = col1;
 						break;
+
 					case 2:
 						tmp = col2;
 						break;
+
 					case 3:
 						tmp = col3;
 						break;
+
 					default:
 						break;
 					}
@@ -888,24 +925,31 @@ static gint32 bitmap_to_gimp(guint8 type)
 			case 0:
 				sb = mcolour0;
 				break;
+
 			case 1:
 				sb = mcolour1;
 				break;
+
 			case 2:
 				sb = mcolour2;
 				break;
+
 			case 3:
 				sb = mcolour3;
 				break;
+
 			case 4:
 				sb = mcolour4;
 				break;
+
 			case 5:
 				sb = mcolour5;
 				break;
+
 			case 6:
 				sb = mcolour6;
 				break;
+
 			case 7:
 				sb = mcolour7;
 				break;
@@ -942,24 +986,31 @@ static gint32 bitmap_to_gimp(guint8 type)
 			case 0:
 				sb = mcolour0;
 				break;
+
 			case 1:
 				sb = mcolour1;
 				break;
+
 			case 2:
 				sb = mcolour2;
 				break;
+
 			case 3:
 				sb = mcolour3;
 				break;
+
 			case 4:
 				sb = mcolour4;
 				break;
+
 			case 5:
 				sb = mcolour5;
 				break;
+
 			case 6:
 				sb = mcolour6;
 				break;
+
 			case 7:
 				sb = mcolour7;
 				break;
@@ -991,12 +1042,15 @@ static gint32 bitmap_to_gimp(guint8 type)
 			case 0:
 				src = (row % 2) ? sprmap0 : sprmap1;
 				break;
+
 			case 1:
 				src = (row % 2) ? sprmap2 : sprmap3;
 				break;
+
 			default:
 				break;
 			}
+
 			src += (row % 42) / 2 * 3;
 
 			for (col = 0; col < 24; col++) {
@@ -1020,14 +1074,17 @@ static gint32 bitmap_to_gimp(guint8 type)
 							*(dst + i) = *scolour1;
 							*(dst + i + 1) = *scolour1;
 							break;
+
 						case 0x2:
 							*(dst + i) = *scolour2;
 							*(dst + i + 1) = *scolour2;
 							break;
+
 						case 0x3:
 							*(dst + i) = *scolour3;
 							*(dst + i + 1) = *scolour3;
 							break;
+
 						case 0x0:
 						default:
 							*(dst + i) = 0xf;
@@ -1056,7 +1113,7 @@ static gint32 bitmap_to_gimp(guint8 type)
 	return image_ID;
 }
 
-static guint8 get_bgcol(gint line, guint8 type)
+static gint8 get_bgcol(gint line, guint8 type)
 {
 	gint col[16][2];
 	gint x, y;
@@ -1080,10 +1137,10 @@ static guint8 get_bgcol(gint line, guint8 type)
 	return col[0][0];
 }
 
-static gboolean get_fgcols(guint8 *col0, guint8 *col1, guint8 *col2,
+static gboolean get_fgcols(gint8 *col0, gint8 *col1, gint8 *col2,
 			   guint8 type, guint x, guint y, guint8 line)
 {
-	guint8 tcol[8][4];
+	gint8 tcol[8][4];
 	guint8 i, j;
 	guint16 cbits[8];
 	guint8 tmp;
@@ -1182,6 +1239,61 @@ static gint gimp_to_rain(gint32 image_ID)
 	return size;
 }
 
+/**
+ * Set a pixel in a 64 bitmap picture
+ *
+ * @param bitmap The bitmap picture to set the pixel in
+ * @param x The x-coordinate of the pixel
+ * @param y The y-coordinate of the pixel
+ */
+static void set_bitmap_pixel(guint8 **bitmap, guint x, guint y)
+{
+	*(*bitmap +
+	  (y / 8) * 320 +
+	  (y % 8) +
+	  (x / 8) * 8) |= (1 << (7 - (x % 8)));
+}
+
+/**
+ * Set a pixel in a 64 spritemap
+ *
+ * @param bitmap The bitmap picture to set the pixel in
+ * @param x The x-coordinate of the pixel
+ * @param y The y-coordinate of the pixel
+ * @param width The width of the sprite map (in pixels)
+ */
+static void set_sprmap_pixel(guint8 **sprmap, guint x, guint y, guint width)
+{
+	*(*sprmap +
+	  (y / 21) * width / 24 * 64 +
+	  (y % 21) * 3 +
+	  (x / 24) * 64 +
+	  (x % 24) / 8) |= (1 << (7 - (x % 8)));
+}
+
+#if 0
+/**
+ * Calculate colour statistics
+ * Note: for multicolour images, this is going to be off by a factor of 2,
+ * but since that holds equally for all colours, this does not matter
+ *
+ * @param cmap The image map to scan
+ * @param colstats An pointer to the an array to store colour stats in
+ */
+static void get_colour_statistics(guchar *cmap, guint8 **colstats)
+{
+	int i;
+
+	/* Empty colstats */
+	for (i = 0; i < 16; i++)
+		**colstats = 0;
+
+	/* Collect the colour statistics */
+	for (i = 0; i < (320 * 200); i++)
+		*(*colstats + cmap[i]) += 1;
+}
+#endif
+
 static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 {
 	GimpDrawable *drawable = NULL;
@@ -1191,11 +1303,13 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 	guint8 *mcolp = NULL;
 	guint8 *colp = NULL;
 	guint8 *bitp;
+	guint8 *sprmap1;
+	guint8 *sprmap2;
 	guchar *cmap;
 	guint8 *bgp = NULL;
 	guint height;
 	guint width;
-	guint ncols;
+	gint ncols;
 	guint size;
 	guint x, y;
 
@@ -1225,7 +1339,13 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 
 	cmap = gimp_image_get_cmap(image_ID, &ncols);
 
-	if (ncols > 16) {
+	if (type == HWS) {
+		if (ncols > 4) {
+			gimp_message("Image has more than 4 colours. "
+				     "Aborting.");
+			return -1;
+		}
+	} else if (ncols > 16) {
 		gimp_message("Image has more than 16 colours. Aborting.");
 		return -1;
 	}
@@ -1256,6 +1376,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + ADVOCP_BG;
 		size = ADVOCP_SIZE;
 		break;
+
 	case AFLI:
 		data[0] = AFLI_BASE & 0xff;
 		data[1] = AFLI_BASE >> 8 & 0xff;
@@ -1264,6 +1385,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		size = AFLI_SIZE;
 		mode = HIRES;
 		break;
+
 	case ARTIST:
 		data[0] = ARTIST_BASE & 0xff;
 		data[1] = ARTIST_BASE >> 8 & 0xff;
@@ -1273,6 +1395,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + ARTIST_BG;
 		size = ARTIST_SIZE;
 		break;
+
 	case BLAZE:
 		data[0] = BLAZE_BASE & 0xff;
 		data[1] = BLAZE_BASE >> 8 & 0xff;
@@ -1282,6 +1405,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = NULL;
 		size = BLAZE_SIZE;
 		break;
+
 	case DRAZ:
 		data[0] = DRAZ_BASE & 0xff;
 		data[1] = DRAZ_BASE >> 8 & 0xff;
@@ -1291,6 +1415,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + DRAZ_BG;
 		size = DRAZ_SIZE;
 		break;
+
 	case FLI:
 		/* Only FLIv2 is supported for saving */
 		data[0] = FLIV2_BASE & 0xff;
@@ -1301,6 +1426,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + FLIV2_BG;
 		size = FLIV2_SIZE;
 		break;
+
 	case IMSYS:
 		data[0] = IMSYS_BASE & 0xff;
 		data[1] = IMSYS_BASE >> 8 & 0xff;
@@ -1310,6 +1436,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = NULL;
 		size = IMSYS_SIZE;
 		break;
+
 	case IP64H:
 		data[0] = IP64H_BASE & 0xff;
 		data[1] = IP64H_BASE >> 8 & 0xff;
@@ -1318,6 +1445,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		size = IP64H_SIZE;
 		mode = HIRES;
 		break;
+
 	case IP64L:
 		data[0] = IP64L_BASE & 0xff;
 		data[1] = IP64L_BASE >> 8 & 0xff;
@@ -1327,6 +1455,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + IP64L_BG;
 		size = IP64L_SIZE;
 		break;
+
 	case KOALA:
 		data[0] = KOALA_BASE & 0xff;
 		data[1] = KOALA_BASE >> 8 & 0xff;
@@ -1336,6 +1465,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + KOALA_BG;
 		size = KOALA_SIZE;
 		break;
+
 	case OCP:
 		data[0] = OCP_BASE & 0xff;
 		data[1] = OCP_BASE >> 8 & 0xff;
@@ -1344,12 +1474,14 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		size = OCP_SIZE;
 		mode = HIRES;
 		break;
+
 	case SPRITE:
 		data[0] = SPRITE_BASE & 0xff;
 		data[1] = SPRITE_BASE >> 8 & 0xff;
 		bitp = data + SPRITE_BITMAP;
 		size = width / 24 * height / 21 * 64 + 2;
 		break;
+
 	case SUXX:
 		/* Only SUXXv3 is supported for saving */
 		data[0] = SUXXV3_BASE & 0xff;
@@ -1360,6 +1492,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + SUXXV3_BG;
 		size = SUXXV3_SIZE;
 		break;
+
 	case TLIGHT:
 		data[0] = TLIGHT_BASE & 0xff;
 		data[1] = TLIGHT_BASE >> 8 & 0xff;
@@ -1369,6 +1502,7 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = data + TLIGHT_BG;
 		size = TLIGHT_SIZE;
 		break;
+
 	case VIDCOM:
 		data[0] = VIDCOM_BASE & 0xff;
 		data[1] = VIDCOM_BASE >> 8 & 0xff;
@@ -1378,6 +1512,17 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		bgp = NULL;
 		size = VIDCOM_SIZE;
 		break;
+
+	case HWS:
+		data[0] = HWS_BASE & 0xff;
+		data[1] = HWS_BASE >> 8 & 0xff;
+		bitp = data + HWS_BITMAP;
+		sprmap1 = data + HWS_SPRMAP1;
+		sprmap2 = data + HWS_SPRMAP2;
+		size = HWS_SIZE;
+		mode = HIRES;
+		break;
+
 	default:
 		gimp_message("Target format not supported. Aborting.");
 		return -1;
@@ -1394,7 +1539,8 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		}
 
 		if (mode == HIRES && type != SPRITE) {
-			gimp_message("Too high resolution. Aborting.");
+			gimp_message("Too high resolution: Aborting.");
+			printf("x: %d, y: %d\n", i % 320, i / 320);
 			return -1;
 		}
 	}
@@ -1479,6 +1625,70 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 				bitp++;
 			}
 		}
+	} else if (type == HWS) {
+		gint col0 = -1, col1 = -1, col2 = -1, col3 = -1;
+
+		/* Find out what colours to use for the bitmap */
+		for (y = 0; y < 200; y++) {
+			for (x = 96; x < 320; x++) {
+				static gint tmp;
+
+				tmp = buf[x + 320 * y] & 0xf;
+
+				if (col0 == tmp || col0 == -1) {
+					col0 = tmp;
+				} else if (col1 == tmp || col1 == -1) {
+					col1 = tmp;
+				} else {
+					gimp_message("Columns 96-319 "
+						     "cannot "
+						     "contain more than 2 "
+						     "colours. Aborting.");
+					return -1;
+				}
+			}
+		}
+
+		for (y = 0; y < 200; y++) {
+			for (x = 0; x < 320; x++) {
+				static gint tmp;
+
+				tmp = buf[x + 320 * y] & 0xf;
+
+				if (col0 == tmp) {
+				} else if (col1 == tmp || col1 == -1) {
+					col1 = tmp;
+					/* set to bitmap */
+					set_bitmap_pixel(&bitp, x, y);
+				} else if (col2 == tmp || col2 == -1) {
+					if (y < 8 || y > 196) {
+						gimp_message("Row 0-7 and "
+							     "197-199 cannot "
+							     "contain more "
+							     "than 2 colours. "
+							     "Aborting.");
+						return -1;
+					}
+
+					col2 = tmp;
+					set_sprmap_pixel(&sprmap1,
+							 x, y - 8, 96);
+				} else {
+					if (y < 8 || y > 196) {
+						gimp_message("Row 0-7 and "
+							     "197-199 cannot "
+							     "contain more "
+							     "than 2 colours. "
+							     "Aborting.");
+						return -1;
+					}
+
+					col3 = tmp;
+					set_sprmap_pixel(&sprmap2,
+							 x, y - 8, 96);
+				}
+			}
+		}
 	} else if (mode == HIRES) {
 		guint8 block[8];
 
@@ -1572,16 +1782,48 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		}
 	} else if (mode == LORES) {
 		guint8 block[8];
+		gint8 col0 = -1, col1 = -1, col2 = -1, bgcol = -1;
+		gboolean is4col = TRUE;
+		guint i;
+
+		/* If the picture contains a maximum of 4 colours,
+		 * always use the same bit pattern for the same colour
+		 */
+		for (i = 0; i < (320 * 200); i += 2) {
+			guint8 tmp;
+
+			tmp = buf[i];
+
+			if (col0 == tmp || col0 == -1) {
+				col0 = tmp;
+			} else if (col1 == tmp || col1 == -1) {
+				col1 = tmp;
+			} else if (col2 == tmp || col2 == -1) {
+				col2 = tmp;
+			} else if (bgcol == tmp || bgcol == -1) {
+				bgcol = tmp;
+			} else {
+				is4col = FALSE;
+				break;
+			}
+		}
 
 		if (bgp) {
 			guint8 bg = 0;
 			guint i;
 
 			for (i = 0; i < 200; i++) {
-				if (type == FLI || i == 0)
+				if (is4col == TRUE) {
+					bg = bgcol;
+				} else if (type == FLI || i == 0)
 					bg = get_bgcol(i, type);
 
-				*(bgp + 24 - (i / 8) + 25 * (i % 8)) = bg;
+				if (type == FLI)
+					*(bgp + 24 - (i / 8) +
+					  25 * (i % 8)) = bg;
+				else
+					*bgp = bg;
+
 				*(tmpbg + i) = bg;
 			}
 		}
@@ -1591,17 +1833,21 @@ static gint gimp_to_bitmap(gint32 image_ID, guint8 type)
 		 * plus background per unit
 		 */
 		for (y = 0; y < 25; y++) {
-			static guint8 col0, col1, col2;
-
 			for (x = 0; x < 40; x++) {
-				static guint8 i, j;
+				static gint j;
 
 				for (i = 0; i < 8; i++) {
 					static gint off = 0;
 
-					if (!get_fgcols(&col0, &col1, &col2,
-							type, x, y, i)) {
+					/* If is4col is TRUE we already
+					 * know what colours to use
+					 */
+					if (is4col == TRUE) {
+					} else if (!get_fgcols(&col0, &col1,
+							       &col2, type,
+							       x, y, i)) {
 						gimp_message("Too many colours in a 4x8 block. Aborting.");
+						printf("x: %d, y: %d\n", x, y);
 						return -1;
 					}
 
